"""
OAuth2 token-based permissions for Django REST Framework views.

"""
import logging

from django.conf import settings
from rest_framework import permissions
from rest_framework.exceptions import ValidationError

from . import lookup

LOG = logging.getLogger(__name__)


def OrPermission(*args):
    """
    This is a function posing as a class. An example of it's intended usage is

    ..code::
        permission_classes = (OrPermission(A, B), )

    where A & B are both class that extend from BasePermission. The function returns a class that,
    when instantiated authorise a request when either A or B authorise that request.

    :param args: a tuple of BasePermission subclasses
    :return: an OrPermissionClass closure
    """
    class OrPermissionClass(permissions.BasePermission):

        def __init__(self):
            # instantiate each permission class
            self.permissions = [Permission() for Permission in args]
            # check that all given permissions inherit from BasePermission
            for permission in self.permissions:
                assert issubclass(type(permission), permissions.BasePermission)

        def has_permission(self, request, view):
            """is true when has_permission() for any of the permissions is true"""
            for permission in self.permissions:
                if permission.has_permission(request, view):
                    return True
            return False

        def has_object_permission(self, request, view, obj):
            """This is true when BOTH has_permission() AND has_object_permission()
            for any of the permissions is true.
            This is important because of the case where X.has_permission() returns false
            but X.has_object_permission() has been left unimplemented and thus returns true.
            This in combination with Y.has_permission() = true and
            Y.has_object_permission() = false would otherwise return true incorrectly.
            """
            for permission in self.permissions:
                if permission.has_permission(request, view) \
                        and permission.has_object_permission(request, view, obj):
                    return True
            return False

    return OrPermissionClass


def AndPermission(*args):
    """
    This is a function posing as a class. An example of it's intended usage is

    ..code::
        permission_classes = (OrPermission(AndPermission(A, B), C), )

    where A & B are both class that extend from BasePermission. The function returns a class that,
    when instantiated authorise a request when both A and B authorise that request.

    :param args: a tuple of BasePermission subclasses
    :return: an AndPermissionClass closure
    """
    class AndPermissionClass(permissions.BasePermission):

        def __init__(self):
            # instantiate each permission class
            self.permissions = [Permission() for Permission in args]
            # check that all given permissions inherit from BasePermission
            for permission in self.permissions:
                assert issubclass(type(permission), permissions.BasePermission)

        def has_permission(self, request, view):
            """is false when has_permission() for any of the permissions is false"""
            for permission in self.permissions:
                if not permission.has_permission(request, view):
                    return False
            return True

        def has_object_permission(self, request, view, obj):
            """is false when has_object_permission() for any of the permissions is false"""
            for permission in self.permissions:
                if not permission.has_object_permission(request, view, obj):
                    return False
            return True

    return AndPermissionClass


class UserInInstitutionPermission(permissions.BasePermission):
    """
    Django REST framework permission which requires that the user acting on an asset be associated
    with the department that the asset belongs to.
    """
    def has_permission(self, request, view):
        """
        When a new asset is created check that the user is associated with the given department.
        """

        # PUT, PATCH, and DELETE should return true here
        # otherwise has_object_permission() will never be called.
        if request.method != 'POST':
            return True

        department = request.data.get('department')
        if not department:
            raise ValidationError('department is required')

        return self._validate_asset_user_institution(request.user, department)

    def has_object_permission(self, request, view, obj):
        """
        When a new asset is changed/deleted
        check that the user is associated with the department of the existing asset
        AND (in the case of changed) is also associated with the given department.
        """
        if request.method in permissions.SAFE_METHODS:
            return True

        if not self._validate_asset_user_institution(request.user, obj.department):
            return False
        # in the case of PATCH, department may not have have been given
        if 'department' in request.data and not \
                self._validate_asset_user_institution(request.user, request.data['department']):
            return False
        return True

    @staticmethod
    def _validate_asset_user_institution(user, department):
        """Validates that the user is member of the department that the asset belongs to
        (asset_department)."""
        if user.is_anonymous:
            return False
        return department in lookup.get_instids_for_email(user.email)


class UserInIARGroupPermission(permissions.BasePermission):
    """
    Django REST framework permission which requires that the user be in the IAR_USERS_LOOKUP_GROUP.

    """
    def has_permission(self, request, view):
        if request.user.is_anonymous:
            return False

        user_group_names = lookup.get_group_names_for_email(request.user.email)
        return settings.IAR_USERS_LOOKUP_GROUP in user_group_names
